import luigi
from luigi.contrib.external_program import ExternalProgramTask
import os

class IndexBam(ExternalProgramTask):
    bam_file = luigi.Parameter()

    def program_args(self):
        return ['samtools', 'index', self.bam_file]

    def output(self):
        return luigi.LocalTarget('{}.bai'.format(self.bam_file))



import datetime
from os.path import join

import luigi

from ..config import bioluigi
from ..scheduled_external_program import ScheduledExternalProgramTask

cfg = bioluigi()

class GenerateReport(ScheduledExternalProgramTask):
    task_namespace = 'multiqc'

    input_dirs = luigi.ListParameter()
    output_dir = luigi.Parameter()

    title = luigi.OptionalParameter(default=None, positional=False)
    comment = luigi.OptionalParameter(default=None, positional=False)

    def program_args(self):
        args = [cfg.multiqc_bin, '--outdir', self.output_dir]
        if self.title is not None:
            args.extend(['--title', self.title])
        if self.comment is not None:
            args.extend(['--comment', self.comment])
        args.extend(self.input_dirs)
        return args

    def output(self):
        return luigi.LocalTarget(join(self.output_dir, 'multiqc_report.html'))



"""
Collection of utilities and mixins for tasks
"""

import logging

import luigi
from luigi.task import getpaths, flatten, flatten_output

logger = logging.getLogger('luigi-interface')

class DynamicWrapperTask(luigi.Task):
    """
    Similar to luigi.task.WrapperTask but for dynamic dependencies yielded in
    the body of the run() method.
    """
    def complete(self):
        # ensure that static dependencies are met
        if not all(req.complete() for req in flatten(self.requires())):
            return False

        # ensure that all dynamic dependencies are met
        try:
            return all(req.complete() for chunk in self.run()
                       for req in flatten(chunk))
        except:
            logger.exception('%s failed at run() step; the exception will not be raised because Luigi is still building the graph.', repr(self))
            return False

class TaskWithOutputMixin(object):
    """
    Extends a luigi.WrapperTask to forward its dependencies as output.
    """
    def output(self):
        return getpaths(self.requires())

class DynamicTaskWithOutputMixin(object):
    """
    Extends a task to forward its dynamic dependencies as output.
    """
    def output(self):
        tasks = []
        if all(req.complete() for req in flatten(self.requires())):
            try:
                tasks = list(self.run())
            except:
                logger.exception('%s failed at run() step; the exception will not be raised because Luigi is still building the graph.', repr(self))

        # FIXME: conserve task structure: the generator actually create an
        # implicit array level even if a single task is yielded.
        # For now, we just handle the special singleton case.
        if len(tasks) == 1:
            tasks = tasks[0]

        return getpaths(tasks)

class CreateTaskOutputDirectoriesBeforeRunMixin(object):
    """
    Ensures that output directories exist before running the task.
    """
    def run(self):
        for out in flatten_output(self):
            out.makedirs()
        return super(CreateTaskOutputDirectoriesBeforeRunMixin, self).run()

class RemoveTaskOutputOnFailureMixin(object):
    """
    Remove a task outputs on failure.

    This only applies for output that have a defined 'remove' method.
    """
    def on_failure(self, err):
        logger.info('Removing task output of %s due to failure.', repr(self))
        for out in flatten_output(self):
            if out.exists() and hasattr(out, 'remove'):
                try:
                    out.remove()
                except:
                    logger.exception('Failed to remove output %s while cleaning up %s.', repr(out), repr(self))
        return super(RemoveTaskOutputOnFailureMixin, self).on_failure(err)



from bioluigi.scheduled_external_program import ScheduledExternalProgramTask
import luigi

class Index(ScheduledExternalProgramTask):
    """Index a tabular format (VCF, TSV, etc.) using tabix utility"""
    input_file = luigi.Parameter()

    preset = luigi.ChoiceParameter(choices=['gff', 'bed', 'sam', 'vcf', 'psltbl'], positional=False)

    seq = luigi.IntParameter(positional=False)
    begin = luigi.IntParameter(positional=False)
    end = luigi.IntParameter(positional=False)
    skip_lines = luigi.IntParameter(positional=False, default=0)
    comment = luigi.Parameter(positional=False)

    def program_args(self):
        args = ['tabix']
        if self.preset:
            args.extend(['-p', self.preset])
        else:
            args.extend([
                '-s', self.seq,
                '-b', self.begin,
                '-e', self.end,
                '-c', self.comment,
                '-S', self.skip_lines])
        args.append(self.input_file)
        return args

    def output(self):
        return luigi.LocalTarget(self.input_file + '.tbi')



import datetime
from os.path import join, basename, splitext
import luigi
from ..scheduled_external_program import ScheduledExternalProgramTask
from ..config import bioluigi

cfg = bioluigi()

class GenerateReport(ScheduledExternalProgramTask):
    task_namespace = 'fastqc'

    input_file = luigi.Parameter()
    output_dir = luigi.Parameter()

    @staticmethod
    def gen_report_basename(fastq_path):
        sample_name, ext = splitext(basename(fastq_path))
        if ext == '.gz':
            sample_name, ext = splitext(sample_name)
        return '{}_fastqc.html'.format(sample_name)

    def program_args(self):
        return [cfg.fastqc_bin,
                '--outdir', self.output_dir,
                self.input_file]

    def output(self):
        return luigi.LocalTarget(join(self.output_dir, self.gen_report_basename(self.input_file)))



import datetime
import os
from os.path import join, split, basename
from tempfile import mkdtemp
import shutil

import luigi
from luigi.task import flatten

from ..scheduled_external_program import ScheduledExternalProgramTask
from ..config import bioluigi

cfg = bioluigi()

class Prefetch(ScheduledExternalProgramTask):
    """
    Prefetch a SRA run archive.

    The number of concurrent prefetch jobs can be adjusted by setting the
    'prefetch_jobs' resource.
    """
    task_namespace = 'sratoolkit'

    srr_accession = luigi.Parameter()
    output_file = luigi.Parameter()
    max_size = luigi.IntParameter(default=20, positional=False, significant=False, description='Maximum download size in gigabytes')
    extra_args = luigi.ListParameter(default=[], positional=False, description='Extra arguments to pass to prefetch which can be used to setup Aspera')

    @property
    def resources(self):
        r = super(Prefetch, self).resources
        r.update({'prefetch_jobs': 1})
        return r

    def program_args(self):
        args = [cfg.prefetch_bin,
                '--max-size',    '{}G'.format(self.max_size),
                '--output-file', self.output().path]

        args.extend(self.extra_args)
        args.append(self.srr_accession)

        return args

    def output(self):
        return luigi.LocalTarget(self.output_file)

class FastqDump(ScheduledExternalProgramTask):
    """
    Extract FASTQs from a SRA archive

    The task output three targets: single-end reads (1 file) and paired-end
    reads (2 files). Note that it is sufficient for either the single-end file
    or both paired-end files to exist for the task to be considered complete.

    The number of concurrent fastq-dump jobs can be adjusted by setting the
    'fastq_dump_jobs' resource.
    """
    task_namespace = 'sratoolkit'

    input_file = luigi.Parameter(description='A file path or a SRA archive, or a SRA run accession')
    output_dir = luigi.Parameter(description='Destination directory for the extracted FASTQs')

    minimum_read_length = luigi.IntParameter(default=0, positional=False, description='Minimum read length to be extracted from the archive')

    @property
    def resources(self):
        r = super(FastqDump, self).resources
        r.update({'fastq_dump_jobs': 1})
        return r

    def __init__(self, *kwargs, **kwds):
        super(FastqDump, self).__init__(*kwargs, **kwds)
        base, tail = split(self.output_dir)
        self.temp_output_dir = mkdtemp(prefix=tail + '-tmp', dir=base)

    def program_args(self):
        args = [cfg.fastqdump_bin,
                '--gzip',
                '--clip',
                '--skip-technical',
                '--readids',
                '--dumpbase',
                '--split-e',
                '--keep-empty-files']

        if self.minimum_read_length > 0:
            args.extend(['-M', self.minimum_read_length])

        args.extend(['--outdir', self.temp_output_dir])

        args.append(self.input_file)

        return args

    def run(self):
        try:
            super(FastqDump, self).run()
            # move every output to the final directory
            for out in self.output():
                tmp_out_path = join(self.temp_output_dir, basename(out.path))
                if os.path.exists(tmp_out_path):
                    out.makedirs()
                    os.replace(tmp_out_path, out.path)
        finally:
            shutil.rmtree(self.temp_output_dir)

    def output(self):
        sra_accession, _ = os.path.splitext(os.path.basename(self.input_file))
        return [luigi.LocalTarget(join(self.output_dir, sra_accession + '.fastq.gz')),
                luigi.LocalTarget(join(self.output_dir, sra_accession + '_1.fastq.gz')),
                luigi.LocalTarget(join(self.output_dir, sra_accession + '_2.fastq.gz'))]

    def complete(self):
        se, r1, r2 = self.output()
        return se.exists() or (r1.exists() and r2.exists())



from os.path import exists, join

import luigi
from luigi.util import requires

from ..config import bioluigi
from ..scheduled_external_program import ScheduledExternalProgramTask

cfg = bioluigi()

class RsemReference(luigi.Target):
    """
    Represents the target of rsem-prepare-reference script.
    """
    def __init__(self, reference_name):
        self.reference_name = reference_name

    def exists(self):
        exts = ['grp', 'ti', 'seq', 'chrlist']
        return all(exists('{}.{}'.format(self.reference_name, ext))
                for ext in exts)

class PrepareReference(ScheduledExternalProgramTask):
    task_namespace = 'rsem'

    annotation_file = luigi.Parameter()
    reference_fasta_files = luigi.ListParameter()
    reference_name = luigi.Parameter()

    aligner = luigi.ChoiceParameter(choices=['star'], positional=False)

    star_path = luigi.OptionalParameter(default=None, positional=False)

    def program_args(self):
        args = [join(cfg.rsem_dir, 'rsem-prepare-reference')]

        args.extend(['--gtf', self.annotation_file])

        if self.aligner == 'star':
            args.append('--star')

            if self.star_path is not None:
                args.extend(['--star-path', self.star_path])

        args.extend(['-p', self.cpus])

        args.extend(self.reference_fasta_files)

        args.append(self.reference_name)

        return args

    def output(self):
        return RsemReference(self.reference_name)

@requires(PrepareReference)
class CalculateExpression(ScheduledExternalProgramTask):
    task_namespace = 'rsem'

    upstream_read_files = luigi.ListParameter()
    sample_name = luigi.Parameter()

    strandedness = luigi.ChoiceParameter(default='none', choices=['none', 'forward', 'reverse'], positional=False)

    extra_args = luigi.ListParameter(default=[], positional=False)

    def program_args(self):
        args = [join(cfg.rsem_dir, 'rsem-calculate-expression')]

        args.extend(['-p', self.cpus])

        if len(self.upstream_read_files) == 1:
            pass
        elif len(self.upstream_read_files) == 2:
            args.append('--paired-end')
        else:
            raise ValueError('Unexpected number of mates: {}.'.format(len(self.upstream_read_files)))

        args.extend(['--strandedness', self.strandedness])

        if self.aligner == 'star':
            args.append('--star')

            if self.star_path is not None:
                args.extend(['--star-path', self.star_path])

            if self.upstream_read_files[0].endswith('.gz'):
                args.append('--star-gzipped-read-file')

        args.extend(self.extra_args)

        args.extend(self.upstream_read_files)
        args.extend([self.reference_name, self.sample_name])

        return args

    def output(self):
        return [luigi.LocalTarget('{}.isoforms.results'.format(self.sample_name)),
                luigi.LocalTarget('{}.genes.results'.format(self.sample_name))]




"""
Command-line interface for interacting with Luigi scheduler.
"""

import json
import requests
import datetime
import click
from fnmatch import fnmatch
import sys
from os.path import join
from collections import Counter
from babel.numbers import format_number

class TooManyTasksError(Exception):
    def __init__(self, num_tasks):
        self.num_tasks = num_tasks
    def __str__(self):
        return 'That request would return {} tasks; try filtering by status, glob query or set the --no-limit flag.'.format(format_number(self.num_tasks))

def rpc(scheduler_url, method, **kwargs):
    url = join(scheduler_url, 'api', method)
    payload = {'data': json.dumps(kwargs)}
    res = requests.get(url, params=payload if kwargs else None)
    res.raise_for_status()
    response_data = res.json()['response']
    if 'num_tasks' in response_data:
        raise TooManyTasksError(response_data['num_tasks'])
    return response_data

def task_sort_key(task):
    """Produce a key to sort tasks by relevance."""
    return datetime.datetime.now() - (task['time_running'] if task['status'] == 'RUNNING' else task['last_updated'])

def task_matches(task, task_glob):
    """Match a task against a glob pattern."""
    return task_glob is None or fnmatch(task['name'], task_glob) or fnmatch(task['display_name'], task_glob)

class TaskFormatter(object):
    """Format a task for texutual display."""
    @staticmethod
    def format_status(status):
        tg = {'DONE': 'green',
              'PENDING': 'yellow',
              'RUNNING': 'blue',
              'FAILED': 'red',
              'DISABLED': 'white',
              'UNKNOWN': 'white'}
        return click.style(status, fg=tg[status]) if status in tg else status
    def format(self, task):
        raise NotImplementedError

class InlineTaskFormatter(TaskFormatter):
    """Format a task for inline display."""
    def __init__(self, task_id_width, status_width=17):
        self.task_id_width = task_id_width
        self.status_width = status_width

    def format(self, task):
        if task['status'] == 'RUNNING':
            tr = (datetime.datetime.now() - task['time_running'])
        else:
            tr = task['last_updated']

        return '{id:{id_width}}\t{status:{status_width}}\t{time}\n'.format(
                id=click.style(task['id'], bold=True),
                id_width=self.task_id_width,
                status=self.format_status(task['status']),
                status_width=self.status_width,
                time=tr)

class DetailedTaskFormatter(TaskFormatter):
    """Format a task for detailed multi-line display."""
    @staticmethod
    def format_dl(dl):
        key_fill = max(len(click.style(k)) for k in dl)
        return '\n\t'.join('{:{key_fill}}\t{}'.format(click.style(k, bold=True) + ':', v, key_fill=key_fill) for k, v in dl.items())

    def format(self, task):
        return '''{id}
Status:      \t{status}
Priority:    \t{priority}
Name:        \t{display_name}
Start time:  \t{start_time}
Last updated:\t{last_updated}
Time running:\t{time_running}
Status message:\n\t{status_message}
Parameters:\n\t{params}
Resources:\n\t{resources}
Workers:\n\t{workers}\n\n'''.format(
        id=click.style(task['id'], bold=True),
        display_name=task['display_name'],
        status=self.format_status(task['status']),
        priority=task['priority'],
        status_message=task['status_message'] if task['status_message'] else 'No status message were received for this task.',
        start_time=task['start_time'],
        last_updated=task['last_updated'],
        time_running=(datetime.datetime.now() - task['time_running']) if task['status'] == 'RUNNING' else '',
        params=self.format_dl(task['params']) if task['params'] else 'No parameters were set.',
        resources=self.format_dl(task['resources']) if task['resources'] else 'No resources were requested for the execution of this task.',
        workers='\n\t'.join(task['workers']) if task['workers'] else 'No workers are assigned to this task.')

class TasksSummaryFormatter(object):
    def format(self, tasks):
        count_by_status = Counter()
        for task in tasks:
            count_by_status[task['status']] += 1
        return '\n'.join('{}\t{}'.format(TaskFormatter.format_status(k), v) for k, v in count_by_status.items())

def fix_tasks_dict(tasks):
    for key, t in tasks.items():
        t['id'] = key
        t['start_time'] = t['start_time'] and datetime.datetime.fromtimestamp(t['start_time'])
        t['time_running'] = t['time_running'] and datetime.datetime.fromtimestamp(t['time_running'])
        t['last_updated'] = t['last_updated'] and datetime.datetime.fromtimestamp(t['last_updated'])

@click.group()
@click.option('--scheduler-url', default='http://localhost:8082/')
@click.pass_context
def main(ctx, scheduler_url):
    ctx.obj = {'SCHEDULER_URL': scheduler_url}

@main.command()
@click.argument('task_glob', required=False)
@click.option('--status', multiple=True)
@click.option('--user', multiple=True)
@click.option('--summary', is_flag=True)
@click.option('--detailed', is_flag=True)
@click.option('--no-limit', is_flag=True)
@click.pass_context
def list(ctx, task_glob, status, user, summary, detailed, no_limit):
    """
    List all tasks that match the given pattern and filters.
    """
    scheduler_url = ctx.obj['SCHEDULER_URL']

    search = task_glob.replace('*', '') if task_glob else None

    limit = None if no_limit else 100000

    tasks = {}
    if status:
        for s in status:
            try:
                tasks.update(rpc(scheduler_url, 'task_list', search=search, status=s, limit=limit))
            except TooManyTasksError as e:
                click.echo(e)
                return
    else:
        try:
            tasks.update(rpc(scheduler_url, 'task_list', search=search, limit=limit))
        except TooManyTasksError as e:
            click.echo(e)
            return

    fix_tasks_dict(tasks)

    filtered_tasks = tasks.values()

    # filter by user
    if user:
        filtered_tasks = [task for task in filtered_tasks
                          if any(u in worker for worker in task['workers'] for u in user)]

    filtered_tasks = [task for task in filtered_tasks
                      if task_matches(task, task_glob)]

    if not filtered_tasks:
        click.echo('No task match the provided query.')
        return

    task_id_width = max(len(click.style(task['id'], bold=True)) for task in filtered_tasks)

    if summary:
        formatter = TasksSummaryFormatter()
        click.echo(formatter.format(filtered_tasks))
    else:
        if detailed:
            formatter = DetailedTaskFormatter()
        else:
            formatter = InlineTaskFormatter(task_id_width=task_id_width)

        click.echo_via_pager(formatter.format(t) for t in sorted(filtered_tasks, key=task_sort_key))

@main.command()
def submit(*args):
    """
    Schedule a given task for execution.
    """
    luigi.cmdline.luigi_run(args)

@main.command()
@click.argument('task_id')
@click.pass_context
def show(ctx, task_id):
    """
    Show the details of a specific task given its identifier.
    TASK_ID Task identifier
    """
    scheduler_url = ctx.obj['SCHEDULER_URL']

    tasks = {}
    for status, t in rpc(scheduler_url, 'task_search', task_str=task_id).items():
        tasks.update(t)
    fix_tasks_dict(tasks)

    formatter = DetailedTaskFormatter()

    try:
        click.echo(formatter.format(tasks[task_id]))
    except KeyError:
        click.echo('No such task %s.' % task_id)
        sys.exit(1)

@main.command()
@click.argument('task_id')
@click.option('--recursive', is_flag=True)
@click.pass_context
def reenable(ctx, task_id, recursive):
    """
    Reenable a disabled task.
    """
    scheduler_url = ctx.obj['SCHEDULER_URL']

    toreenable = [task_id]

    if recursive:
        deps = rpc(scheduler_url, 'dep_graph', task_id=task_id)
        toreenable.extend(k for k in deps if deps[k]['status'] == 'DISABLED')

    for task_id in toreenable:
        try:
            rpc(scheduler_url, 're_enable_task', task_id=task_id)
            click.echo('%s has been re-enabled.' % task_id)
        except requests.exceptions.HTTPError as e:
            click.echo('Failed to re-enable {}: {}'.format(task_id, e))
            continue

@main.command()
@click.argument('task_id')
@click.option('--recursive', is_flag=True)
@click.pass_context
def forgive(ctx, task_id, recursive):
    """
    Forgive a failed task.
    """
    scheduler_url = ctx.obj['SCHEDULER_URL']

    toforgive = []

    if recursive:
        deps = rpc(scheduler_url, 'dep_graph', task_id=task_id)
        toforgive.extend(k for k in deps if deps[k]['status'] == 'FAILED')

    for task_id in toforgive:
        try:
            rpc(scheduler_url, 'forgive_failures', task_id=task_id)
            click.echo('%s has been forgiven.' % task_id)
        except requests.exceptions.HTTPError as e:
            click.echo('Failed to forgive {}: {}'.format(task_id, e))
            continue



import luigi
from os.path import join

from ..scheduled_external_program import ScheduledExternalProgramTask
from ..config import bioluigi

cfg = bioluigi()

class BcftoolsTask(ScheduledExternalProgramTask):
    task_namespace = 'bcftools'

    input_file = luigi.Parameter()

    include = luigi.OptionalParameter(positional=False, default=None)
    exclude = luigi.OptionalParameter(positional=False, default=None)
    regions = luigi.ListParameter(default=[], positional=False)
    regions_file = luigi.OptionalParameter(positional=False, default=None)
    samples = luigi.ListParameter(default=[], positional=False)
    samples_file = luigi.OptionalParameter(default=None, positional=False)
    apply_filters = luigi.OptionalParameter(positional=False, default=None)

    # FIXME: the '--threads' flag does not seem to work

    def subcommand_args(self):
        """Returns specific sub-command arguments."""
        raise NotImplementedError

    def subcommand_input_args(self):
        """
        Returns arguments to be appended at the input file location.

        This is meant to be to to deal with commands that accept multiple input
        files.
        """
        return [self.input_file]

    def program_args(self):
        args = [cfg.bcftools_bin]

        args.extend(self.subcommand_args())

        if self.include is not None:
            args.extend(['-i', self.include])

        if self.exclude is not None:
            args.extend(['-e', self.exclude])

        if self.regions:
            args.extend(['-r', ','.join(self.regions)])

        if self.regions_file is not None:
            args.extend(['-R', self.regions_file])

        if self.samples:
            args.extend(['-s', ','.join(self.samples)])

        if self.samples_file:
            args.extend(['-S', self.samples_file])

        if self.apply_filters is not None:
            args.extend(['-f', self.apply_filters])

        args.extend(self.subcommand_input_args())

        return args

class View(BcftoolsTask):
    """
    View a VCF and apply transformations and filters defined in :class:`BcftoolsTask`.
    """
    output_file = luigi.Parameter()
    output_format = luigi.Parameter(positional=False, default='z')

    def subcommand_args(self):
        return ['view',
            '--output-type', self.output_format,
            '--output-file', self.output_file]

    def output(self):
        return luigi.LocalTarget(self.output_file)

class Annotate(BcftoolsTask):
    """
    Annotate a VCF using bcftools annotate.
    """
    output_file = luigi.Parameter()
    output_format = luigi.Parameter(positional=False, default='z')

    # options given an annotation file
    annotations_file = luigi.OptionalParameter(positional=False, default=None)

    columns = luigi.ListParameter(positional=False, default=[])

    rename_chrs = luigi.OptionalParameter(positional=False, default=None)

    def subcommand_args(self):
        args = ['annotate']

        if self.rename_chrs is not None:
            args.extend(['--rename-chrs', self.rename_chrs])

        if self.annotations_file:
            args.extend(['-a', self.annotations_file])
            args.extend(['-c', ','.join(self.columns)])

        args.extend([
            '--output-type', self.output_format,
            '--output', self.output_file])

        return args

    def output(self):
        return luigi.LocalTarget(self.output_file)

class Sort(BcftoolsTask):
    output_file = luigi.Parameter()
    output_format = luigi.Parameter(positional=False, default='z')

    tmp_dir = luigi.Parameter(default='/tmp', significant=False)

    def subcommand_args(self):
        return ['sort',
                '--temp-dir', self.tmp_dir,
                '--output-type', self.output_format,
                '--output', self.output_file]

    def output(self):
        return luigi.LocalTarget(self.output_file)

class Index(BcftoolsTask):
    """
    Use tabix to create a tabular index for a VCF.
    """
    def subcommand_args(self):
        return ['index', '--tbi']

    def output(self):
        return luigi.LocalTarget(self.input_file + '.tbi')

class Intersect(BcftoolsTask):
    input_file2 = luigi.Parameter()
    output_dir = luigi.Parameter()

    def subcommand_args(self):
        return ['isec', '-p', self.output_dir]

    def subcommand_input_args(self):
        return [self.input_file, self.input_file2]

    def output(self):
        return [luigi.LocalTarget(join(self.output_dir, '000{}.vcf.gz'.format(i))) for i in range(4)]

class Merge(BcftoolsTask):
    """
    Merge the samples of two or more VCF files
    """
    input_file = luigi.ListParameter()

    filter_logic = luigi.ChoiceParameter(default='+', choices=['x', '+'], positional=False)
    info_rules = luigi.ListParameter(default=[], positional=False)

    output_file = luigi.Parameter()
    output_format = luigi.Parameter(positional=False, default='z')

    def subcommand_args(self):
        args = ['merge']

        args.extend(['--filter-logic', self.filter_logic])

        if self.info_rules:
            args.extend(['--info-rules', ','.join(self.info_rules)])

        args.extend([
            '--output-type', self.output_format,
            '--output', self.output_file])

        return args

    def subcommand_input_args(self):
        return self.input_file

    def output(self):
        return luigi.LocalTarget(self.output_file)



import datetime
import logging

import luigi
from luigi.task import flatten_output

from ..scheduled_external_program import ScheduledExternalProgramTask
from ..config import bioluigi
from .utils import RemoveTaskOutputOnFailureMixin

logger = logging.getLogger('luigi-interface')

cfg = bioluigi()

class CutadaptTask(RemoveTaskOutputOnFailureMixin, ScheduledExternalProgramTask):
    """
    Base class for all cutadapt-derived tasks.
    """
    task_namespace = 'cutadapt'

    adapter_3prime = luigi.OptionalParameter(default='', positional=False)
    adapter_5prime = luigi.OptionalParameter(default='', positional=False)

    cut = luigi.IntParameter(default=0, positional=False)
    trim_n = luigi.BoolParameter(default=False, positional=False)
    minimum_length = luigi.IntParameter(default=0, positional=False)

    def program_args(self):
        args = [cfg.cutadapt_bin]

        args.extend(['-j', self.cpus])

        if self.adapter_3prime:
            args.extend(['-a', self.adapter_3prime])

        if self.adapter_5prime:
            args.extend(['-g', self.adapter_5prime])

        if self.cut:
            args.extend(['-u', self.cut])

        if self.trim_n:
            args.append('--trim-n')

        if self.minimum_length:
            args.extend(['--minimum-length', self.minimum_length])

        return args

class TrimReads(CutadaptTask):
    """
    For consistency with TrimPairedReads, this task output a list with a single
    target corresponding to trimmed FASTQ.
    """
    input_file =  luigi.Parameter()
    output_file = luigi.Parameter()

    def program_args(self):
        args = super(TrimReads, self).program_args()
        args.extend(['-o', self.output_file, self.input_file])
        return args

    def output(self):
        return [luigi.LocalTarget(self.output_file)]

class TrimPairedReads(CutadaptTask):
    input_file = luigi.Parameter()
    input2_file = luigi.Parameter()
    output_file = luigi.Parameter()
    output2_file = luigi.Parameter()

    reverse_adapter_3prime = luigi.OptionalParameter(default='', positional=False)
    reverse_adapter_5prime = luigi.OptionalParameter(default='', positional=False)

    def program_args(self):
        args = super(TrimPairedReads, self).program_args()
        if self.reverse_adapter_3prime:
            args.extend(['-A', self.reverse_adapter_3prime])
        if self.reverse_adapter_5prime:
            args.extend(['-G', self.reverse_adapter_5prime])
        args.extend([
            '-o', self.output_file,
            '-p', self.output2_file,
            self.input_file, self.input2_file])
        return args

    def output(self):
        return [luigi.LocalTarget(self.output_file), luigi.LocalTarget(self.output2_file)]



import luigi

class bioluigi(luigi.Config):
    scheduler = luigi.Parameter(default='local', significant=False, description='Default scheduler to use in ScheduledExternalProgram')
    scheduler_partition = luigi.OptionalParameter(default='', significant=False, description='Node partition to use for scheduling jobs if supported')
    scheduler_extra_args = luigi.ListParameter(default=[], significant=False, description='List of extra arguments to pass to the scheduler')
    prefetch_bin = luigi.Parameter(default='prefetch')
    fastqdump_bin = luigi.Parameter(default='fastq-dump')
    cutadapt_bin = luigi.Parameter(default='cutadapt')
    fastqc_bin = luigi.Parameter(default='fastqc')
    star_bin = luigi.Parameter(default='STAR')
    rsem_dir = luigi.Parameter(default='')
    bcftools_bin = luigi.Parameter(default='bcftools')
    vep_bin = luigi.Parameter(default='vep')
    vep_dir = luigi.OptionalParameter(default='', significant=False)
    multiqc_bin = luigi.Parameter(default='multiqc')



import luigi
from os.path import join

from ..config import bioluigi
from ..scheduled_external_program import ScheduledExternalProgramTask

cfg = bioluigi()

class Annotate(ScheduledExternalProgramTask):
    """
    Annotate a VCF file with Ensembl VEP.

    Setting :cpus: parameter will adjust the number of forks (via --fork) VEP
    uses to annotate its input.
    """
    task_namespace = 'vep'

    vcf_file = luigi.Parameter()
    annotated_vcf_file = luigi.Parameter()

    cache = luigi.BoolParameter(default=False, positional=False, significant=False)
    offline = luigi.BoolParameter(default=False, positional=False, significant=False)

    buffer_size = luigi.IntParameter(default=5000, positional=False, significant=False)

    species = luigi.Parameter(positional=False)
    assembly = luigi.Parameter(positional=False)

    plugins = luigi.ListParameter(default=[], positional=False)

    extra_args = luigi.ListParameter(default=[])

    output_format = luigi.ChoiceParameter(choices=['vcf', 'tab', 'json'], default='vcf')
    compress_output = luigi.ChoiceParameter(choices=['gzip', 'bgzip'], default='bgzip')

    def program_args(self):
        args = [cfg.vep_bin,
                '-i',                self.vcf_file,
                '--format',          'vcf',
                '--fork',            self.cpus,
                '--buffer_size',     self.buffer_size,
                '--species',         self.species,
                '--assembly',        self.assembly,
                f'--{self.output_format}',
                '--compress_output', self.compress_output,
                '--output_file',     self.output().path]

        if self.cache:
            args.append('--cache')

        if self.offline:
            args.append('--offline')

        if cfg.vep_dir is not None:
            args.extend(['--dir', cfg.vep_dir])

        for plugin in self.plugins:
            args.extend(['--plugin', plugin])

        args.extend(self.extra_args)

        return args

    def output(self):
        return luigi.LocalTarget(self.annotated_vcf_file)



"""
This module provides a flavour of :class:`luigi.contrib.external_program.ExternalProgramTask`
designed to work with an external scheduler such as Slurm.

It also extend the basic definition of an external task to declare basic
resource consumptions such as the number of CPU and the amount of memory
required to execute the task.
"""

import luigi
from luigi.contrib.external_program import ExternalProgramTask, ExternalProgramRunError, ExternalProgramRunContext
from subprocess import Popen, PIPE, check_call
import os
import datetime
import logging

from .config import bioluigi

cfg = bioluigi()

logger = logging.getLogger('luigi-interface')

_schedulers = {}

def register_scheduler(blurb):
    global _schedulers
    def wrapper(cls):
        _schedulers[blurb] = cls
        return cls
    return wrapper

class Scheduler(object):
    """
    :param blurb: Short name by with the scheduler is referred to
    or if it should be done through Luigi's resource management system.
    """

    @classmethod
    def run_task(self, task):
        raise NotImplementedError

@register_scheduler('slurm')
class SlurmScheduler(Scheduler):
    """
    Scheduler based on Slurm https://slurm.schedmd.com/
    """
    @classmethod
    def run_task(self, task):
        secs = int(task.walltime.total_seconds())
        srun_args = ['srun']
        srun_args.extend([
            '--job-name', repr(task),
            '--time', '{}-{:02d}:{:02d}:{:02d}'.format(secs // 86400, (secs % 86400) // 3600, (secs % 3600) // 60, secs % 60),
            '--mem', '{}G'.format(int(task.memory)),
            '--cpus-per-task', str(task.cpus)])
        if cfg.scheduler_partition:
            srun_args.extend(['--partition', cfg.scheduler_partition])
        # FIXME: task.priority is not reliable and does not reflect what the
        # scheduler
        # TODO: srun_args.extend([--priority', str(max(0, cfg.scheduler_priority))])
        srun_args.extend(map(str, task.scheduler_extra_args))
        args = list(map(str, task.program_args()))
        env = task.program_environment()
        logger.info('Running Slurm command {}'.format(' '.join(srun_args + args)))
        proc = Popen(srun_args + args, env=env, stdout=PIPE, stderr=PIPE, universal_newlines=True)
        with ExternalProgramRunContext(proc):
            stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            raise ExternalProgramRunError('Program exited with non-zero return code.', args, env, stdout, stderr)
        if task.capture_output:
            logger.info('Program stdout:\n{}'.format(stdout))
            logger.info('Program stderr:\n{}'.format(stderr))

class ScheduledExternalProgramTask(ExternalProgramTask):
    """
    Variant of :class:`luigi.contrib.external_program.ExternalProgramTask` that
    executes the task with a :class:`Scheduler`.
    """
    scheduler = luigi.ChoiceParameter(default=cfg.scheduler, choices=['local'] + [blurb for blurb in _schedulers], positional=False, significant=False, description='Scheduler to use for running the task')
    scheduler_partition = luigi.OptionalParameter(default=cfg.scheduler_partition, positional=False, significant=False, description='Scheduler partition (or queue) to use if supported')
    scheduler_extra_args = luigi.ListParameter(default=cfg.scheduler_extra_args, positional=False, significant=False, description='Extra arguments to pass to the scheduler')

    walltime = luigi.TimeDeltaParameter(default=datetime.timedelta(), positional=False, significant=False, description='Amout of time to allocate for the task, default value of zero implies unlimited time')
    cpus = luigi.IntParameter(default=1, positional=False, significant=False, description='Number of CPUs to allocate for the task')
    memory = luigi.FloatParameter(default=1, positional=False, significant=False, description='Amount of memory (in gigabyte) to allocate for the task')

    def __init__(self, *kwargs, **kwds):
        super(ScheduledExternalProgramTask, self).__init__(*kwargs, **kwds)
        try:
            if self.scheduler != 'local':
                self._scheduler = _schedulers[self.scheduler]
        except KeyError:
            raise ValueError('Unsupported scheduler {}'.format(self.scheduler))

    @property
    def resources(self):
        if self.scheduler == 'local':
            # local_jobs is actually constrained by the number of workers
            return {'cpus': self.cpus, 'memory': self.memory}
        else:
            return {'{}_jobs'.format(self.scheduler): 1}

    def run(self):
        if self.scheduler == 'local':
            return super(ScheduledExternalProgramTask, self).run()
        else:
            return self._scheduler.run_task(self)



import luigi

from ..scheduled_external_program import ScheduledExternalProgramTask

luigi.namespace('bioluigi.mirbooking', scope=__name__)

class Mirbooking(ScheduledExternalProgram):
    """
    Solve the microtargetome equilibrium given a set of targets RNAs and
    microRNAs
    """
    input_file = luigi.Parameter()
    targets_file = luigi.Parameter()
    mirnas_file = luigi.Parameter()

    cutoff = luigi.FloatParameter(default=100)

    def program_environment(self):
        env = super(Mirbooking, self).program_environment()
        env.update({'OMP_NUM_THREADS': self.cpus})
        return env

    def program_args(self):
        return ['mirbooking',
                '--targets', self.targets_file,
                '--mirnas',  self.mirnas_file,
                '--input',   self.input_file,
                '--cutoff',  self.cutoff]



from setuptools import setup, find_packages

setup(name='bioluigi',
      version='0.0.4',
      description='Reusable and maintained Luigi tasks to incorporate in bioinformatics pipelines',
      long_description='file: README.md',
      author='Guillaume Poirier-Morency',
      author_email='poirigui@msl.ubc.ca',
      packages=find_packages(),
      entry_points={
          'console_scripts': [
              'bioluigi = bioluigi.cli:main']},
      install_requires=['click', 'luigi', 'requests', 'babel'])



import luigi
from bioluigi.scheduled_external_program import ScheduledExternalProgramTask
import bioluigi.config
import urllib

cfg = bioluigi.config.bioluigi()

class GenerateIndex(ScheduledExternalProgramTask):
    genome_fastas = luigi.ListParameter()
    sjdb = luigi.Parameter()

    output_dir = luigi.Parameter()

    def program_args(self):
        return [cfg.star_bin,
                '--runCommand', 'generateGenome',
                '--genomeDir', self.output_dir]

class GenerateIndexFromEnsembl(luigi.Task):
    taxon = luigi.Parameter()
    version = luigi.IntParameter()

    def run(self):
        urllib.urlretrieve('ftp://', '')
        urllib.urlretrieve('ftp://', '')
        yield GenerateIndex()

class Align(ScheduledExternalProgramTask):
    """
    The task output the alignment and the splice junctions.

    :stranded: Whether the reads are strand-specific
    """
    fastqs = luigi.ListParameter()
    genome_dir = luigi.Parameter()
    output_dir = luigi.Parameter()

    output_format = luigi.ChoiceParameter(choices=['sam', 'bam'], default='sam')
    sort_output = luigi.BoolParameter()
    gzipped_reads = luigi.BoolParameter(default=True)
    stranded_reads = luigi.BoolParameter(default=False)

    # performance feature
    use_shared_memory = luigi.BoolParameter(default=True)

    cpu = 1
    mem = 32

    def program_args(self):
        args = [cfg.star_bin,
                '--runThreadN', self.resources['cpu'],
                '--genomeDir', self.genome_dir]

        if self.stranded_reads:
            args.extend(['--strand', 'Stranded'])

        if self.gzipped_reads:
            args.extend(['--readFileCommand', 'zcat'])

        if len(self.fastqs) == 1:
            pass
        elif len(self.fastqs) == 2:
            pass
        else:
            raise ValueError('')

        args.append('--readFilesIn')
        args.extend(self.fastqs)

        if self.use_shared_memory:
            args.extend(['--genomeLoad', 'LoadAndRemove'])

        return args

    def output(self):
        return [luigi.LocalTarget(join(self.output_dir, 'Aligned.{}.out'.format(self.output_format))),
                luigi.LocalTarget(join(self.output_dir, 'SJ.tab.out'))]
